export { logWithViteTag };
export { logWithVikeTag };
export { logDirectly };
export { isFirstLog };
export { clearScreen };
export { screenHasErrors };
import { assert, projectInfo, stripAnsi, hasProp, assertIsNotProductionRuntime } from '../../utils.js';
import pc from '@brillout/picocolors';
import { isErrorDebug } from '../isErrorDebug.js';
import { getViteDevServer } from '../../../runtime/globalContext.js';
assertIsNotProductionRuntime();
let isFirstLog = true;
let screenHasErrors = false;
function logWithVikeTag(msg, logType, category, showVikeVersion = false) {
    const projectTag = getProjectTag(showVikeVersion);
    msg = prependTags(msg, projectTag, category, logType);
    logDirectly(msg, logType);
}
function getProjectTag(showVikeVersion) {
    let projectTag;
    if (showVikeVersion) {
        projectTag = `[${projectInfo.projectName}@${projectInfo.projectVersion}]`;
    }
    else {
        projectTag = `[${projectInfo.projectName}]`;
    }
    return projectTag;
}
function logWithViteTag(msg, logType, category) {
    msg = prependTags(msg, '[vite]', category, logType);
    logDirectly(msg, logType);
}
// Not production => every log is triggered by logDirectly()
//  - Even all Vite logs also go through logDirectly() (see interceptors of loggerVite.ts)
//  - Production => logs aren't managed by loggerNotProd.ts => logDirectly() is never called (not even loaded as asserted by assertIsVitePluginCode())
function logDirectly(thing, logType) {
    applyViteSourceMapToStackTrace(thing);
    isFirstLog = false;
    if (logType === 'info') {
        console.log(thing);
        return;
    }
    if (logType === 'warn') {
        console.warn(thing);
        return;
    }
    if (logType === 'error') {
        screenHasErrors = true;
        console.error(thing);
        return;
    }
    if (logType === 'error-recover') {
        // stderr because user will most likely want to know about error recovering
        console.error(thing);
        return;
    }
    assert(false);
}
function clearScreen(viteConfig) {
    // We use Vite's logger in order to respect the user's `clearScreen: false` setting
    viteConfig.logger.clearScreen('error');
    screenHasErrors = false;
}
function applyViteSourceMapToStackTrace(thing) {
    if (isErrorDebug())
        return;
    if (!hasProp(thing, 'stack'))
        return;
    const viteDevServer = getViteDevServer();
    if (!viteDevServer)
        return;
    // Apply Vite's source maps
    viteDevServer.ssrFixStacktrace(thing);
}
function prependTags(msg, projectTag, category, logType) {
    const color = (s) => {
        if (logType === 'error' && !hasRed(msg))
            return pc.bold(pc.red(s));
        if (logType === 'error-recover' && !hasGreen(msg))
            return pc.bold(pc.green(s));
        if (logType === 'warn' && !hasYellow(msg))
            return pc.yellow(s);
        if (projectTag === '[vite]')
            return pc.bold(pc.cyan(s));
        if (projectTag.startsWith(`[${projectInfo.projectName}`))
            return pc.bold(pc.cyan(s));
        assert(false);
    };
    let tag = color(`${projectTag}`);
    if (category) {
        tag = tag + pc.dim(`[${category}]`);
    }
    const timestamp = pc.dim(new Date().toLocaleTimeString());
    const whitespace = /\s|\[/.test(stripAnsi(msg)[0]) ? '' : ' ';
    return `${timestamp} ${tag}${whitespace}${msg}`;
}
function hasRed(str) {
    // https://github.com/brillout/picocolors/blob/e291f2a3e3251a7f218ab6369ae94434d85d0eb0/picocolors.js#L57
    return str.includes('\x1b[31m');
}
function hasGreen(str) {
    // https://github.com/brillout/picocolors/blob/e291f2a3e3251a7f218ab6369ae94434d85d0eb0/picocolors.js#L58
    return str.includes('\x1b[32m');
}
function hasYellow(str) {
    // https://github.com/brillout/picocolors/blob/e291f2a3e3251a7f218ab6369ae94434d85d0eb0/picocolors.js#L59
    return str.includes('\x1b[33m');
}
