export { serializePageContextClientSide };
export { serializePageContextAbort };
import { stringify } from '@brillout/json-serializer/stringify';
import { assert, assertWarning, hasProp, unique } from '../utils.js';
import { isErrorPage } from '../../../shared/error-page.js';
import { addIs404ToPageProps } from '../../../shared/addIs404ToPageProps.js';
import pc from '@brillout/picocolors';
import { notSerializable } from '../../../shared/notSerializable.js';
const PASS_TO_CLIENT = [
    'abortReason',
    '_urlRewrite',
    '_urlRedirect',
    'abortStatusCode',
    '_abortCall',
    /* Not needed on the client-side
    '_abortCaller',
    */
    '_pageContextInitHasClientData',
    '_pageId'
];
const PASS_TO_CLIENT_ERROR_PAGE = ['pageProps', 'is404', '_isError'];
function serializePageContextClientSide(pageContext) {
    const passToClient = getPassToClient(pageContext);
    const pageContextClient = {};
    passToClient.forEach((prop) => {
        // We set non-existing props to `undefined`, in order to pass the list of passToClient values to the client-side
        pageContextClient[prop] = pageContext[prop];
    });
    if (Object.keys(pageContext._pageContextInit).some((p) => passToClient.includes(p))) {
        pageContextClient._pageContextInitHasClientData = true;
    }
    let pageContextSerialized;
    try {
        pageContextSerialized = serialize(pageContextClient);
    }
    catch (err) {
        const h = (s) => pc.cyan(s);
        let hasWarned = false;
        const propsNonSerializable = [];
        passToClient.forEach((prop) => {
            const propName = JSON.stringify(prop);
            const varName = h(`pageContext[${propName}]`);
            try {
                serialize(pageContext[prop], varName);
            }
            catch (err) {
                hasWarned = true;
                propsNonSerializable.push(prop);
                assert(hasProp(err, 'messageCore', 'string'));
                assertWarning(false, [
                    `${varName} cannot be serialized and, therefore, cannot be passed to the client.`,
                    `Make sure that ${varName} is serializable, or remove ${h(propName)} from ${h('passToClient')}.`,
                    `Serialization error: ${err.messageCore}.`
                ].join(' '), { onlyOnce: false });
            }
        });
        assert(hasWarned);
        propsNonSerializable.forEach((prop) => {
            pageContextClient[prop] = notSerializable;
        });
        pageContextSerialized = serialize(pageContextClient);
    }
    return pageContextSerialized;
}
function serialize(value, varName) {
    return stringify(value, { forbidReactElements: true, valueName: varName });
}
function getPassToClient(pageContext) {
    let passToClient = [...pageContext._passToClient, ...PASS_TO_CLIENT];
    if (isErrorPage(pageContext._pageId, pageContext._pageConfigs)) {
        assert(hasProp(pageContext, 'is404', 'boolean'));
        addIs404ToPageProps(pageContext);
        passToClient.push(...PASS_TO_CLIENT_ERROR_PAGE);
    }
    passToClient = unique(passToClient);
    return passToClient;
}
function serializePageContextAbort(pageContext) {
    assert(pageContext._urlRedirect || pageContext._urlRewrite || pageContext.abortStatusCode);
    assert(pageContext._abortCall);
    assert(pageContext._abortCaller);
    // Not needed on the client-side
    delete pageContext._abortCaller;
    const unknownProps = Object.keys(pageContext).filter((prop) => ![
        // prettier-ignore
        '_abortCall',
        /* Not needed on the client-side
        '_abortCaller',
        */
        '_urlRedirect',
        '_urlRewrite',
        'abortStatusCode',
        'abortReason',
        'is404',
        'pageProps'
    ].includes(prop));
    if (!pageContext._isLegacyRenderErrorPage) {
        assert(unknownProps.length === 0);
    }
    else {
        // TODO/v1-release: remove
        assertWarning(unknownProps.length === 0, [
            "The following pageContext values won't be available on the client-side:",
            unknownProps.map((p) => `  pageContext[${JSON.stringify(p)}]`),
            'Use `throw render()` instead of `throw RenderErrorPage()`'
        ].join('\n'), {
            onlyOnce: false
        });
    }
    return serialize(pageContext);
}
