export { renderPageAlreadyRouted };
export { prerenderPage };
export { prerender404Page };
export { getPageContextInitEnhanced };
export { getRenderContext };
import { getErrorPageId } from '../../../shared/error-page.js';
import { getHtmlString } from '../html/renderHtml.js';
import { getPageFilesAll } from '../../../shared/getPageFiles.js';
import { assert, assertUsage, hasProp, isNotNullish, objectAssign, unique } from '../utils.js';
import { serializePageContextClientSide } from '../html/serializePageContextClientSide.js';
import { addUrlComputedProps } from '../../../shared/addUrlComputedProps.js';
import { getGlobalContext } from '../globalContext.js';
import { createHttpResponseObject, createHttpResponsePageContextJson } from './createHttpResponseObject.js';
import { loadPageFilesServerSide } from './loadPageFilesServerSide.js';
import { executeOnRenderHtmlHook } from './executeOnRenderHtmlHook.js';
import { executeOnBeforeRenderHooks } from './executeOnBeforeRenderHook.js';
import { logRuntimeError } from './loggerRuntime.js';
import { isNewError } from './isNewError.js';
import { preparePageContextForUserConsumptionServerSide } from './preparePageContextForUserConsumptionServerSide.js';
import { executeGuardHook } from '../../../shared/route/executeGuardHook.js';
import { loadPageRoutes } from '../../../shared/route/loadPageRoutes.js';
import pc from '@brillout/picocolors';
async function renderPageAlreadyRouted(pageContext) {
    // pageContext._pageId can either be the:
    //  - ID of the page matching the routing, or the
    //  - ID of the error page `_error.page.js`.
    assert(hasProp(pageContext, '_pageId', 'string'));
    const isError = pageContext.is404 || !!pageContext.errorWhileRendering;
    assert(isError === (pageContext._pageId === getErrorPageId(pageContext._pageFilesAll, pageContext._pageConfigs)));
    objectAssign(pageContext, await loadPageFilesServerSide(pageContext));
    if (!isError) {
        await executeGuardHook(pageContext, (pageContext) => preparePageContextForUserConsumptionServerSide(pageContext));
    }
    if (!isError) {
        await executeOnBeforeRenderHooks(pageContext);
    }
    else {
        try {
            await executeOnBeforeRenderHooks(pageContext);
        }
        catch (err) {
            if (isNewError(err, pageContext.errorWhileRendering)) {
                logRuntimeError(err, pageContext._httpRequestId);
            }
        }
    }
    if (pageContext.isClientSideNavigation) {
        if (isError) {
            objectAssign(pageContext, { _isError: true });
        }
        const pageContextSerialized = serializePageContextClientSide(pageContext);
        const httpResponse = await createHttpResponsePageContextJson(pageContextSerialized);
        objectAssign(pageContext, { httpResponse });
        return pageContext;
    }
    const renderHookResult = await executeOnRenderHtmlHook(pageContext);
    if (renderHookResult.htmlRender === null) {
        objectAssign(pageContext, { httpResponse: null });
        return pageContext;
    }
    else {
        const { htmlRender, renderHook } = renderHookResult;
        const httpResponse = await createHttpResponseObject(htmlRender, renderHook, pageContext);
        objectAssign(pageContext, { httpResponse });
        return pageContext;
    }
}
async function prerenderPage(pageContext) {
    objectAssign(pageContext, {
        isClientSideNavigation: false,
        _urlHandler: null
    });
    /* Should we execute the guard() hook upon pre-rendering? Is there a use case for this?
     *  - It isn't trivial to implement, as it requires to duplicate / factor out the isAbortError() handling
    await executeGuardHook(pageContext, (pageContext) => preparePageContextForUserConsumptionServerSide(pageContext))
    */
    await executeOnBeforeRenderHooks(pageContext);
    const { htmlRender, renderHook } = await executeOnRenderHtmlHook(pageContext);
    assertUsage(htmlRender !== null, `Cannot pre-render ${pc.cyan(pageContext.urlOriginal)} because the ${renderHook.hookName}() hook defined by ${renderHook.hookFilePath} didn't return an HTML string.`);
    assert(pageContext.isClientSideNavigation === false);
    const documentHtml = await getHtmlString(htmlRender);
    assert(typeof documentHtml === 'string');
    if (!pageContext._usesClientRouter) {
        return { documentHtml, pageContextSerialized: null, pageContext };
    }
    else {
        const pageContextSerialized = serializePageContextClientSide(pageContext);
        return { documentHtml, pageContextSerialized, pageContext };
    }
}
async function prerender404Page(renderContext, pageContextInit_) {
    const errorPageId = getErrorPageId(renderContext.pageFilesAll, renderContext.pageConfigs);
    if (!errorPageId) {
        return null;
    }
    const pageContext = {
        _pageId: errorPageId,
        _httpRequestId: null,
        _urlRewrite: null,
        is404: true,
        routeParams: {},
        // `prerender404Page()` is about generating `dist/client/404.html` for static hosts; there is no Client Routing.
        _usesClientRouter: false,
        _routeMatches: []
    };
    const pageContextInit = {
        urlOriginal: '/fake-404-url',
        ...pageContextInit_
    };
    {
        const pageContextInitEnhanced = getPageContextInitEnhanced(pageContextInit, renderContext);
        objectAssign(pageContext, pageContextInitEnhanced);
    }
    objectAssign(pageContext, await loadPageFilesServerSide(pageContext));
    return prerenderPage(pageContext);
}
function getPageContextInitEnhanced(pageContextInit, renderContext, { urlComputedPropsNonEnumerable = false, ssr: { urlRewrite, urlHandler, isClientSideNavigation } = {
    urlRewrite: null,
    urlHandler: null,
    isClientSideNavigation: false
} } = {}) {
    assert(pageContextInit.urlOriginal);
    const globalContext = getGlobalContext();
    const pageContextInitEnhanced = {
        ...pageContextInit,
        _objectCreatedByVitePluginSsr: true,
        // The following is defined on `pageContext` because we can eventually make these non-global (e.g. sot that two pages can have different includeAssetsImportedByServer settings)
        _baseServer: globalContext.baseServer,
        _baseAssets: globalContext.baseAssets,
        _includeAssetsImportedByServer: globalContext.includeAssetsImportedByServer,
        // TODO: use GloablContext instead
        _pageFilesAll: renderContext.pageFilesAll,
        _pageConfigs: renderContext.pageConfigs,
        _pageConfigGlobal: renderContext.pageConfigGlobal,
        _allPageIds: renderContext.allPageIds,
        _pageRoutes: renderContext.pageRoutes,
        _onBeforeRouteHook: renderContext.onBeforeRouteHook,
        _pageContextInit: pageContextInit,
        _urlRewrite: urlRewrite,
        _urlHandler: urlHandler,
        isClientSideNavigation
    };
    addUrlComputedProps(pageContextInitEnhanced, !urlComputedPropsNonEnumerable);
    return pageContextInitEnhanced;
}
// TODO: remove getRenderContext() in favor of getGlobalObject() + reloadGlobalContext()
// TODO: impl GlobalNodeContext + GlobalClientContext + GloablContext, and use GlobalContext instead of RenderContext
async function getRenderContext() {
    const globalContext = getGlobalContext();
    const { pageFilesAll, allPageIds, pageConfigs, pageConfigGlobal } = await getPageFilesAll(false, globalContext.isProduction);
    const { pageRoutes, onBeforeRouteHook } = await loadPageRoutes(pageFilesAll, pageConfigs, pageConfigGlobal, allPageIds);
    assertNonMixedDesign(pageFilesAll, pageConfigs);
    const renderContext = {
        pageFilesAll: pageFilesAll,
        pageConfigs,
        pageConfigGlobal,
        allPageIds: allPageIds,
        pageRoutes,
        onBeforeRouteHook
    };
    return renderContext;
}
function assertNonMixedDesign(pageFilesAll, pageConfigs) {
    if (pageFilesAll.length === 0 || pageConfigs.length === 0)
        return;
    const indent = '- ';
    const v1Files = unique(pageConfigs
        .map((p) => Object.values(p.configValues)
        .map(({ definedAtInfo }) => definedAtInfo)
        .filter(isNotNullish)
        .map((definedAtInfo) => indent + definedAtInfo.filePath))
        .flat(2));
    assertUsage(false, [
        'Mixing the new V1 design with the old V0.4 design is forbidden.',
        'V1 files:',
        ...v1Files,
        'V0.4 files:',
        ...pageFilesAll.map((p) => indent + p.filePath)
    ].join('\n'));
}
